#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License,
    or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, see <http://www.gnu.org/licenses/>.

    @author: mkaay, RaNaN
"""

import re
import sys
import tempfile
import os

from os import listdir
from os.path import isfile, join, exists, abspath
from shutil import copy
from sys import version_info
from itertools import chain
from traceback import print_exc

from module.lib.SafeEval import const_eval as literal_eval
from module.ConfigParser import IGNORE_PLUGINS, IGNORE_ACCOUNTS
from module.utils import SYNO_PYLOAD_CONF_DIR_PATH

ACC_VERSION = 1

class PluginManager:
    ROOT = "module.plugins."
    TYPES = ("crypter", "container", "hoster", "accounts")

    PATTERN = re.compile(r'__pattern__.*=.*r("|\')([^"\']+)')
    VERSION = re.compile(r'__version__.*=.*("|\')([0-9.]+)')
    CONFIG = re.compile(r'__config__.*=.*(\[[^\]]+\])', re.MULTILINE)
    DESC = re.compile(r'__description__.?=.?("|"""|\')([^"\']+)')
    MULTIHOSTER = re.compile(r'class [\w]+\(MultiAccount\):')
    accounts = dict()
    premiumSupportList = [
        "FilefactoryCom",
        "FourSharedCom",
        "LinksnappyCom",
        "RapidgatorNet",
        "RealdebridCom",
        "UploadedTo",
        "UptoboxCom"
    ]

    def __init__(self, method):

        if method in ('listAccountPlugin', 'verifyAccount', 'setAccount', 'checkURL'):
            self.loadAccounts()
        self.plugins = {}
        self.createIndex(method)


    def createIndex(self, method):
        """create information for all plugins available"""

        sys.path.append(abspath(""))
        if method in ('checkURL'):
            self.crypterPlugins, temp_config = self.parse("crypter", pattern=True)
            self.plugins["crypter"] = self.crypterPlugins
            self.hosterPlugins, temp_config = self.parse("hoster", pattern=True)
            self.plugins["hoster"] = self.hosterPlugins
            self.containerPlugins, temp_config = self.parse("container", pattern=True)
            self.plugins["container"] = self.containerPlugins
            self.accountPlugins, temp_config = self.parse("accounts")
            self.plugins["accounts"] = self.accountPlugins
        if method in ('listAccountPlugin', 'verifyAccount'):
            self.accountPlugins, temp_config = self.parse("accounts")
            self.plugins["accounts"] = self.accountPlugins


    def parse(self, folder, pattern=False, home={}):
        """
        returns dict with information
        home contains parsed plugins from module.

        {
        name : {path, version, config, (pattern, re), (plugin, class)}
        }

        """
        plugins = {}
        pfolder = join(pypath, "module", "plugins", folder)

        configs = {}
        for f in listdir(pfolder):
            if (isfile(join(pfolder, f)) and f.endswith(".py") or f.endswith("_25.pyc") or
                    f.endswith("_26.pyc") or f.endswith("_27.pyc")) and not f.startswith("_"):
                data = open(join(pfolder, f))
                content = data.read()
                data.close()

                if f.endswith("_25.pyc") and version_info[0:2] != (2, 5):
                    continue
                elif f.endswith("_26.pyc") and version_info[0:2] != (2, 6):
                    continue
                elif f.endswith("_27.pyc") and version_info[0:2] != (2, 7):
                    continue

                name = f[:-3]
                if name[-1] == ".": name = name[:-4]

                version = self.VERSION.findall(content)
                if version:
                    version = version[0][1]
                else:
                    version = 0

                if name in IGNORE_PLUGINS or (folder, name) in IGNORE_PLUGINS:
                     continue

                plugins[name] = {}
                plugins[name]["version"] = version

                module = f.replace(".pyc", "").replace(".py", "")

                plugins[name]["pluginname"] = module

                if module in self.accounts:
                    if 0 < len(self.accounts[module].keys()):
                        plugins[name]["login"] = self.accounts[module].keys()[0]

                if pattern:
                    pattern = self.PATTERN.findall(content)

                    if pattern:
                        pattern = pattern[0][1]
                    else:
                        pattern = "^unmachtable$"

                    plugins[name]["pattern"] = pattern

                    try:
                        plugins[name]["re"] = re.compile(pattern)
                    except:
                        pass

                if folder == "accounts":
                    m = self.MULTIHOSTER.findall(content)
                    plugins[name]["bl_multihoster"] = (0 != len(m))

                # internals have no config
                if folder == "internal":
                    continue

                config = self.CONFIG.findall(content)
                if config:
                    config = literal_eval(config[0].strip().replace("\n", "").replace("\r", ""))
                    desc = self.DESC.findall(content)
                    desc = desc[0][1] if desc else ""

                    if type(config) == list and all(type(c) == tuple for c in config):
                        config = dict((x[0], x[1:]) for x in config)
                    else:
                        continue

                    if folder == "hooks" and "activated" not in config:
                        config['activated'] = ["bool", "Activated", False]

                    config['desc'] = desc
                    configs[name] = config

        if not home:
            temp_plugins, temp_configs = self.parse(folder, pattern, plugins or True)
            plugins.update(temp_plugins)
            configs.update(temp_configs)

        return plugins, configs


    def parseUrls(self, urls):
        """parse plugins for given list of urls"""

        last = None
        res = [] # tupels of (url, plugin)

        for url in urls:
            if type(url) not in (str, unicode, buffer): continue

            if last and last[1]["re"].match(url):
                res.append((url, last[0]))
                continue

            for name, value in chain(self.crypterPlugins.iteritems(), self.hosterPlugins.iteritems(), self.containerPlugins.iteritems()):
                if value["re"].match(url):
                    res.append((url, name))
                    last = (name, value)
                    break
            for name, value in self.accountPlugins.iteritems():
                if not value["bl_multihoster"]:
                    continue
                if "login" not in value:
                    continue
                username = self.accounts[name].keys()[0]
                password = self.accounts[name][username]["password"]
                try:
                    module = self.transform_module(self.ROOT + "accounts.%s" % (value["pluginname"]))
                    hoster = module[value["pluginname"]](None)
                    pattern = hoster.get_pattern(username, password)
                    if not pattern:
                        continue
                    if re.compile(pattern).match(url):
                        res.append((url, name))
                        break
                except:
                    continue
        return res

    def findAccountPlugin(self, name):
        if name in self.plugins['accounts']:
            return self.plugins['accounts'][name]
        return None

    def transform_module(self, namespace):
        path = namespace.replace(".", "/") + ".py"
        path = join(pypath, path)
        module = {}

        with open(path, 'r') as read_file:
            content = read_file.read()
            content = content.replace('from ...', 'from module.')
            content = content.replace('from ..', 'from module.plugins.')
            content = content.replace('from .', 'from module.plugins.accounts.')
            content = content.replace('from module.plugins.internal.Account import Account', 'from module.plugins.internal.SynoAccount import Account')
            content = content.replace('from module.plugins.internal.MultiAccount import MultiAccount', 'from module.plugins.internal.SynoMultiAccount import MultiAccount')
            content = content.replace('from module.plugins.internal.XFSAccount import XFSAccount', 'from module.plugins.internal.SynoXFSAccount import XFSAccount')
            exec content in module

        return module

    def loadAccounts(self):
        """loads all accounts available"""

        import synodl
        accounts_conf_file = join(SYNO_PYLOAD_CONF_DIR_PATH, "accounts.conf")
        accounts_conf_file_bak = join(SYNO_PYLOAD_CONF_DIR_PATH, "accounts.backup")

        if not exists(accounts_conf_file):
            f = open(accounts_conf_file, "wb")
            f.write("version: " + str(ACC_VERSION))
            f.close()

        f = open(accounts_conf_file, "rb")
        content = f.readlines()
        version = content[0].split(":")[1].strip() if content else ""
        f.close()

        if not version or int(version) < ACC_VERSION:
            copy(accounts_conf_file, accounts_conf_file_bak)
            f = open(accounts_conf_file, "wb")
            f.write("version: " + str(ACC_VERSION))
            f.close()
            return

        plugin = ""
        name = ""

        for line in content[1:]:
            line = line.strip()

            if not line: continue
            if line.startswith("#"): continue
            if line.startswith("version"): continue

            if line.endswith(":") and line.count(":") == 1:
                plugin = line[:-1]
                self.accounts[plugin] = {}

            elif line.startswith("@"):
                try:
                    option = line[1:].split()
                    self.accounts[plugin][name]["options"][option[0]] = [] if len(option) < 2 else ([option[1]] if len(option) < 3 else option[1:])
                except:
                    pass

            elif ":" in line:
                name, sep, encryptedPW = line.partition(":")
                pw = synodl.Decrypt(encryptedPW)
                self.accounts[plugin][name] = {"password": pw, "options": {}, "valid": True}


    def saveAccounts(self):
        """save all account information"""
        import synodl
        accountConfPath = join(SYNO_PYLOAD_CONF_DIR_PATH, "accounts.conf")
        f = open(accountConfPath, "wb")
        f.write("version: " + str(ACC_VERSION) + "\n")

        for plugin, accounts in self.accounts.iteritems():
            f.write("\n")
            f.write(plugin+":\n")

            for name,data in accounts.iteritems():
                encryptedPW = synodl.Encrypt(data["password"])
                f.write("\n\t%s:%s\n" % (name,encryptedPW) )
                if data["options"]:
                    for option, values in data["options"].iteritems():
                        f.write("\t@%s %s\n" % (option, " ".join(values)))

        f.close()
        os.chmod(f.name, 0600)

    def verifyAccount(self, host, username, password):
        """ Returns loaded module for plugin

        :param type: plugin type, subfolder of module.plugins
        :param name:
        """
        """ If password is None, use saved password for verification if username is the same. """
        if password is None:
            if host not in self.accounts:
                raise Exception("Failed to find saved account info")
            if 0 == len(self.accounts[host].keys()):
                raise Exception("There is no saved account info for this host")
            if username != self.accounts[host].keys()[0]:
                raise Exception("Different username without password")
            data = self.accounts[host][username]
        else:
            data = {'password': password}

        plugin = self.findAccountPlugin(host)

        if None == plugin:
            raise Exception("Failed to find account plugin")

        bl_valid = False;
        bl_premium = False;
        login_data = {}
        try:
            module = self.transform_module(self.ROOT + "accounts.%s" % (plugin["pluginname"]))
            hoster = module[plugin["pluginname"]](None)
            bl_valid = hoster.login(username, data["password"], login_data)
            if bl_valid:
                bl_premium = hoster.premium(username, data["password"], login_data)
        except:
            print_exc()
            raise Exception("Failed to verify account plugin")
        return bl_valid, bl_premium

    def setAccount(self, host, username, password):
        if "" == username and "" == password:
            # clear account information if both username and password are empty
            self.accounts.pop(host, None)
        else:
            # doesn't allow empty password here
            if "" == username or "" == password:
                raise Exception("Username or password is empty")
            record = dict()
            data = dict()
            data['password'] = password
            data['options'] = dict()
            record[username] = data
            self.accounts[host] = record
        self.saveAccounts()
        pass

    def listAccountPlugin(self):
        """return list of account plugin names"""
        from module.ConfigParser import ConfigParser
        c = ConfigParser()
        configData = c.parseConfig("plugin.conf")
        configDataKeys = configData.keys()
        accountPluginArr = []
        for key in self.accountPlugins.keys():
            if key in IGNORE_ACCOUNTS:
                continue
            data = self.accountPlugins[key]
            data["is_tested"] = key in self.premiumSupportList
            data["can_be_disabled"] = False
            if key in configDataKeys:
                if 'use_premium' in configData[key]:
                    data["use_premium"] = configData[key]['use_premium']['value']
                    data["can_be_disabled"] = True
            accountPluginArr.append(self.accountPlugins[key])
        return accountPluginArr

    @staticmethod
    def setEnabled(host, enabled):
        from module.ConfigParser import ConfigParser
        pluginConfigPath = "plugin.conf"
        parser = ConfigParser()
        configData = parser.parseConfig(pluginConfigPath)
        if host not in configData.keys():
            raise Exception("This host can't be found in config")
        if 'use_premium' not in configData[host].keys():
            raise Exception("use_premium can't be set for this host")
        configData[host]['use_premium']['value'] = enabled
        parser.saveConfig(configData, pluginConfigPath)

    def checkURL(self, url):
        urlList = []
        urlList.append(url)
        result = self.parseUrls(urlList)
        if 0 == len(result):
            return False
        else:
            return True
